MODULE OmDerivative_mod
use Dal
use omallocation_mod
IMPLICIT NONE


PUBLIC :: Derivative

interface Derivative
     
    Module procedure Derivative1, Derivative2
end interface
    

CONTAINS
!*************************************************************
!
!
!*************************************************************
real(double) Function Derivative1(n, x, index, h, func, deriv, err)
    integer, intent(in) :: n, index
    real(double) :: x(1 : n)
    real(double), intent(inout) :: h

    real(double), intent(out) :: err
    real(double), parameter :: con = 1.4_double
    real(double), parameter :: con2 = con * con
    real(double), parameter :: big = 1.0e30
    integer, parameter :: ntab = 10
    real(double), parameter :: safe = 2.0_double
    integer :: i, j, iminus1, jminus1
    real(double) :: errt, fac, hh
    real(double) :: a(1 : ntab, 1 : ntab)

    interface
    
    Function func(x) result(f)
    use dal
        real(double) :: f
        real(double) :: x(:)
    end function func
    end interface

    interface

    Function Deriv(n, x, index, h, func) result (f)
    use dal
    real(double) :: f
    integer, intent(in) :: n, index
    real(double) :: x(1 : n)
    real(double), intent(in) :: h

  interface
    
    Function func(x) result(f)
    use dal
        real(double) :: f
        real(double) :: x(:)
    end function func
    end interface


    end function deriv
    end interface


    err = big
    hh = h

!write(*, *) "Derivaitive line 73 n=", n, " index=", index   
    a(1, 1) = deriv(n, x, index, h, func)

    derivative1 = 0.0_double

    do i = 2, ntab
        iminus1 = i - 1
        hh = hh / con
      

        a(1, i) =  deriv(n, x, index, hh, func)
        fac = con2
        do j = 2, i
            jminus1 = j - 1
            a(j, i) = (a(jminus1, i) * fac - a(jminus1, iminus1)) / (fac - 1.0_double)
            fac = con2 * fac
            errt = max(abs(a(j, i) - a(jminus1, i)), abs(a(j, i) - a(jminus1, iminus1)))
            if (errt .lt. err) then
                err = errt
                derivative1 = a(j, i) 
        
            end if
         
        end do
        if (abs(a(i, i) - a(iminus1, iminus1)) .ge. safe * err) then
!             write(*, *) "Exiting"
            exit

        end if
    end do
!if (abs(derivative - 1.0_double) .le. 0.0000001_double) then


!write(*, *) "Error hh=", hh
!write(*, *) "i=", i, " j=", j, " ntab=", ntab
!write(*, *) "a(", j, ",", i, ")=", a(j, i)
!write(*, *) "a(", i, ",", i, ")=", a(i, i)
!write(*, *) "a(", i - 1, ",", i - 1, ")=", a(i - 1, i - 1)


!end if

    
end Function Derivative1
!*****************************************************
!
!***************************************************** 

real(double) Function Derivative2(n, x, index, h, func)
    integer, intent(in) :: n, index
    real(double) :: x(1 : n)
    real(double), intent(in) :: h

    real(double), dimension(1 : n) :: x1, x2
   



    interface
    
    Function func(x) result(f)
    use dal
        real(double) :: f
        real(double) :: x(:)
    end function func
    end interface

 

    x1 = x
    x2 = x 
    x1(index) = x(index) + h
    x2(index) = x(index) - h

!write(*, *) "Derivative2 h=", h, " ubound x=", lbound(x), " ubound=", ubound(x), " index=", index
!write(*, *) "Derivative2 x1=", x1
!write(*, *) "x2=", x2
    Derivative2 = (func(x1) - func(x2)) / (2.0_double * h)
!write(*, *) "Derivative=", Derivative2       

end function Derivative2



!*****************************************************
!
!***************************************************** 

real(double) Function Derivative3(n, x, index, h, func)
    integer, intent(in) :: n, index
    real(double) :: x(1 : n)
    real(double), intent(in) :: h

    real(double), dimension(1 : n) :: x1
    real(double) :: f1, f2, f3, f4   

    interface
    
    Function func(x) result(f)
    use dal
        real(double) :: f
        real(double) :: x(:)
    end function func
    end interface

!write(*, *) "derivative 3 n=", n, " index=", index 
    
    x1(index) = x(index) + 2.0_double * h
  

    f1 = func(x1)

   
    x1(index) = x(index) + h
  
    f2 = func(x1)

       
    x1(index) = x(index) -  h
  
    f3 = func(x1)

    x1(index) = x(index) - 2.0_double * h
  
    f4 = func(x1)

  
!write(*, *) "Derivative2 h=", h, " ubound x=", lbound(x), " ubound=", ubound(x), " index=", index
!write(*, *) "Derivative2 x1=", x1
!write(*, *) "x2=", x2

    Derivative3 = (-f1 + 8.0_double * (f2 - f3) + f4) / (12.0_double * h)


!write(*, *) "Derivative=", Derivative2       

end function Derivative3

end module OmDerivative_mod    








