module ominterpolate_mod

    use param
    use cal
    use dal
    use omallocation_mod
    IMPLICIT NONE

public :: BilinearInterpolate
public :: BilinearInterpolation1, BilinearInterpolation2,  &
          BilinearInterpolation3

interface  BilinearInterpolate
    module procedure  BilinearInterpolation1,  &
                      BilinearInterpolation2,  &
                      BilinearInterpolation3,  &
                      BilinearInterpolation4,  &
                      BilinearInterpolation5
end interface
public :: bilinearConstruct, BilinearDestruct







PUBLIC :: polintconstruct, polintdestruct, polintinterpolate
public :: polintinterpolate1,  polintinterpolate2, polintinterpolate3

interface polintinterpolate
     
    Module procedure  polintinterpolate1,  polintinterpolate2, polintinterpolate3
end interface
    

type BilinearStruct

    integer :: nx = 0, ny = 0
   
    real(single) :: xl = 0.0, xu = 0.0, yl = 0.0, yu = 0.0

    real(single) :: xcentre = 0.0, ycentre = 0.0       
    real(single), pointer :: image(:, :) => null() ! The image


end type BilinearStruct




type polintStruct

    integer :: n
    real(single), pointer :: x(:) => null()
    real(single), pointer :: y(:) => null()
    real(single), pointer :: c(:) => null()
    real(single), pointer :: d(:) => null()

end type polintstruct

contains



subroutine BilinearConstruct(image, bilinearData)
    real(single) :: image(:, :)   
    type(BilinearStruct), intent(inout) :: bilinearData
    integer :: nx, ny

    nx = size(image, 1)
    ny = size(image, 2)
    bilinearData%nx = nx
    bilinearData%ny = ny
   
    bilinearData%xl = 0.0
    bilinearData%yl = 0.0
    bilinearData%xu = real(nx)
    bilinearData%yu = real(ny)
    bilinearData%xcentre = real(nx) / 2.0
    bilinearData%ycentre = real(ny) / 2.0


    bilinearData%image => allocate2dSingle(0, nx - 1, 0, ny - 1, "bilinearData%image")
    bilinearData%image = image

end subroutine BilinearConstruct

subroutine BilinearDestruct(bilinearData)
       
    type(BilinearStruct), intent(inout) :: bilinearData

    if (associated(bilinearData%image)) then
        deallocate(bilinearData%image)
        nullify(bilinearData%image)

    end if
end subroutine BilinearDestruct


subroutine polintConstruct(n, x, y, polintdata)
    integer, intent(in) :: n 
    real(single), dimension(1 : n) :: x, y

    type(polintstruct), intent(out) :: polintdata

    polintdata%n = n
    polintdata%x => allocateSingle(1, n, "x")
    polintdata%x(1 : n) = x(1 : n)
    polintdata%y => allocateSingle(1, n, "y")
    polintdata%y(1 : n) = y(1 : n)


    polintdata%c => allocateSingle(1, n, "c")
    polintdata%d => allocateSingle(1, n, "d")

end subroutine  polintConstruct

subroutine polintDestruct(polintData)
    type(polintstruct), intent(inout) :: polintdata
 
    deallocate(polintdata%x)
    nullify(polintdata%x)
    deallocate(polintdata%y)
    nullify(polintdata%y)



    deallocate(polintdata%c)
    nullify(polintdata%c)
    deallocate(polintdata%d)
    nullify(polintdata%d)
 
end subroutine polintDestruct


!*************************************************
!
!*************************************************
real(single) Function polintInterpolate1(x, polintdata, error)
    real(single), intent(in) :: x
    type(polintstruct), intent(in) :: polintdata
    real(single), intent(out) :: error
    polintInterpolate1 = polintInterpolate3(x, polintData%n, &
         polintData%x, polintData%y, polintData%c, polintData%d, error)

end function  polintInterpolate1
!*************************************************
!
!*************************************************
real(single) Function polintInterpolate2(x, n, xa, ya, error)
    real(single), intent(in) :: x
    integer :: n
    real(single), intent(in) :: xa(1 : n), ya(1 : n)
    real(single), intent(out) :: error
    integer :: i, m, ns = 1
    real(single) :: den, dif, dift, ho, hp, w
    real(single), pointer :: c(:) => null(), d(:) => null()

    dif = abs(x - xa(1))
    c => allocateSingle(1, n, "c")
    d => allocateSingle(1, n, "d")

    ns = 1

    do i = 1, n
        dift = abs(x - xa(i))
        if (dift .lt. dif) then
            ns = i
            dif = dift
        end if

        c(i) = ya(i)
        d(i) = ya(i)
!write(*, *) "Line 123 i=", i, " c=", c(i), " d=", d(i)
    end do

    polintInterpolate2 = ya(ns)
!write(*, *) "Line 127 Polintinterpolate2=", polintinterpolate2
!write(*, *) "xa=", xa
!write(*, *) "ya=", ya

    ns = ns - 1
 
    do m = 1, n - 1
        do i = 1, n - m
            ho = xa(i) - x
            hp = xa(i + m) - x
            w = c(i + 1) - d(i)
            den = ho - hp
            den = w / den
            d(i) = hp * den
            c(i) = ho * den
            write(*, *) "m=", m, " i=", i, " c=", c(i), " d=", d(i), " ns=", ns
        end do

        if ((2 * ns) < (n - m)) then
            error = c(ns + 1)
        else
            error = d(ns)
            ns = ns - 1
        end if
       
        polintInterpolate2 = polintInterpolate2 + error
    end do
!write(*, *) "interpolated value=", polintinterpolate2, " error=", error
    deallocate(d, c)
    nullify(d, c)
call fatal("stopped")
end function polintInterpolate2         

!*************************************************
!
!*************************************************
real(single) Function polintInterpolate3(x, n, xa, ya, c, d, error)
    real(single), intent(in) :: x
    integer :: n
    real(single) :: xa(:), ya(:), c(:), d(:)
    real(single), intent(out) :: error
    integer :: i, m, ns = 1
    real(single) :: den, dif, dift, ho, hp, w
  

    dif = abs(x - xa(1))
 

    ns = 1
    do i = 1, n
        dift = abs(x - xa(i))
        if (dift .lt. dif) then
            ns = i
            dif = dift
        end if
        c(i) = ya(i)
        d(i) = ya(i)
    end do

    polintinterpolate3 = ya(ns)

    ns = ns - 1
    do m = 1, n - 1
        do i = 1, n - m
            ho = xa(i) - x
            hp = xa(i + m) - x
            w = c(i + 1) - d(i)
            den = ho - hp
            den = w / den
            d(i) = hp * den
            c(i) = ho * den
        end do

        if ((2 * ns) < (n - m)) then
            error = c(ns + 1)
        else
            error = d(ns)
            ns = ns - 1
        end if 
        polintInterpolate3 = polintInterpolate3 + error
    end do



end function polintInterpolate3         







!***********************************************************
!
!***********************************************************
real(single) Function  BilinearInterpolation1(imagePtr, x, y)
    real(single), pointer :: imagePtr(:, :)
    real(single), intent(in) :: x, y

    real(single) :: xl, xu, yl, yu
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny

    nx = size(imagePtr, 1)
    ny = size(imagePtr, 2)
    xl = 0.0
    yl = 0.0
    xu = real(nx)
    yu = real(ny)

    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(ny - 1, k0)


        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge. nx) then
            t = 0.0
            j1 = nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if
        if (k1 .ge. ny) then
            k1 = ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if
      
        BilinearInterpolation1 = &
              (1.0 - t) * (1.0 - u) * imagePtr(j0, k0) + &
               t * (1.0 - u) * imagePtr(j1, k0) + &
               t * u * imagePtr(j1, k1) + &
               (1.0  - t) * u * imagePtr(j0, k1)



        

        else
            BilinearInterpolation1 = -1.0
    end if


end function BilinearInterpolation1










!***********************************************************
!
!***********************************************************
real(single) Function BilinearInterpolation2(image, xl, xu, yl, yu, x, y)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: xl, xu, yl, yu, x, y
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny

    nx = size(image, 1)
    ny = size(image, 2)
    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, nint(x) - 1)  ! if x = 3.1 j0 = 2, 2.99 j0 = 1
        j0 = min(nx - 2, j0)
      
        k0 = max(0, nint(y) - 1)
        k0 = min(ny - 2, k0)
        
        j1 = j0 + 1
        k1 = k0 + 1

        t = (x - real(j0) - 0.5)  
        u = (y - real(k0) - 0.5)

        if (t >= 0.0 .and. u >= 0.0) then
            BilinearInterpolation2 = &
              (1.0 - t) * (1.0 - u) * image(j0, k0) + &
               t * (1.0 - u) * image(j1, k0) + &
               t * u * image(j1, k1) + &
               (1.0  - t) * u * image(j0, k1)



        else
            BilinearInterpolation2 = -1.0

        


        end if
    end if


end function BilinearInterpolation2



!***********************************************************
!
!***********************************************************
real(single) Function BilinearInterpolation3(image, dx, dy, xIn, yIn)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: dx, dy, xIn, yIn
  
    real(single) :: x, y, xl, xu, yl, yu
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny, idx, idy
    
    idx = nint(1.0 / dx)
    idy = nint(1.0 / dy)

    x = xIn / dx
    y = Yin / dy
    nx = size(image, 1)
    ny = size(image, 2)

    xl = 0.0
    yl = 0.0
    xu = real(nx - 1)
    yu = real(ny - 1)
    
    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, nint(x) - 1)  ! if x = 3.1 j0 = 2, 2.99 j0 = 1
        j0 = min(nx - 2, j0)
      
        k0 = max(0, nint(y) - 1)
        k0 = min(ny - 2, k0)

       
        j1 = j0 + 1
        k1 = k0 + 1
       
        t = x - real(j0) - 0.5  
        u = y - real(k0) - 0.5

        if (t >= 0.0 .and. u >= 0.0) then
            BilinearInterpolation3 = &
              (1.0 - t) * (1.0 - u) * image(j0, k0) + &
               t * (1.0 - u) * image(j1, k0) + &
               t * u * image(j1, k1) + &
               (1.0  - t) * u * image(j0, k1)


        else
            BilinearInterpolation3 = -1.0

        


        end if
    end if


end function BilinearInterpolation3


!***********************************************************
!
!***********************************************************
real(single) Function  BilinearInterpolation4(bilinearData, x, y)
    type(BilinearStruct), intent(in) :: bilinearData
    
    real(single), intent(in) :: x, y

  
    real(single) :: u, t
    integer j0, k0, j1, k1
    

    if (x .gt. bilinearData%xl .and. x .lt. bilinearData%xu .and. & 
        y .gt. bilinearData%yl .and. y .lt. bilinearData%yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(bilinearData%nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(bilinearData%ny - 1, k0)


        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge.bilinearData%nx) then
            t = 0.0
            j1 = bilinearData%nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if

        if (k1 .ge. bilinearData%ny) then
            k1 = bilinearData%ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if
      
        BilinearInterpolation4 = &
              (1.0 - t) * (1.0 - u) * bilinearData%image(j0, k0) + &
               t * (1.0 - u) * bilinearData%image(j1, k0) + &
               t * u * bilinearData%image(j1, k1) +        &
               (1.0  - t) * u * bilinearData%image(j0, k1)



        

        else
            BilinearInterpolation4 = -1.0
    end if


end function BilinearInterpolation4

!***********************************************************
!
!***********************************************************
real(single) Function  BilinearInterpolation5(dx, dy, bilinearData)
    type(BilinearStruct), intent(in) :: bilinearData
    
    real(single), intent(in) :: dx, dy

  
    real(single) :: u, t, x, y
    integer j0, k0, j1, k1


    x = bilinearData%xcentre + dx
    y = bilinearData%ycentre + dy


    if (x .gt. bilinearData%xl .and. x .lt. bilinearData%xu .and. & 
        y .gt. bilinearData%yl .and. y .lt. bilinearData%yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(bilinearData%nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(bilinearData%ny - 1, k0)


        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge.bilinearData%nx) then
            t = 0.0
            j1 = bilinearData%nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if

        if (k1 .ge. bilinearData%ny) then
            k1 = bilinearData%ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if
      
        BilinearInterpolation5 = &
              (1.0 - t) * (1.0 - u) * bilinearData%image(j0, k0) + &
               t * (1.0 - u) * bilinearData%image(j1, k0) + &
               t * u * bilinearData%image(j1, k1) +        &
               (1.0  - t) * u * bilinearData%image(j0, k1)



        

        else
            BilinearInterpolation5 = -1.0
    end if


end function BilinearInterpolation5


!**************************************************
!
!*************************************************
real(single) Function NeighbourInterpolate(xpos, ypos, n, x, y, f)
    real(single), intent(in) :: xPos, yPos
    integer, intent(in) :: n
    real(single), dimension(1 : n) :: x, y, f
    real(single) :: dx, dy, rsq
    real(double) :: sum1, sum2
    integer :: i

    sum1 = 0.0_double
    sum2 = 0.0_double


    do i = 1, n
        dx = x(i) - xpos
        dy = y(i) - ypos
        rsq = dx * dx + dy * dy
        if(rsq .gt. 1.0e-4) then 
            sum1 = sum1 + f(i) / rsq
            sum2 = sum2 + dble(1.0 / rsq)
        end if  
    end do
    if (sum2 .gt. 1.0e-9) then
        NeighbourInterpolate = real(sum1 / sum2)

    else
        NeighbourInterpolate = 0.0
    end if
end Function NeighbourInterpolate




end module ominterpolate_mod


