!******************************************************************
! Module OmdetectSkybackground_mod
!******************************************************************
module OmSkybackground_mod
    use cal
    use dal     
    use omallocation_mod
    use omnumerics_mod
    use omsourceparam_mod
    use omfunctions_mod
    use brent_mod
    use nonlinls_mod
interface BackgroundMode
    module procedure BackgroundMode1,  &
                     BackgroundMode2,  &
                     ComputeBackgroundMode,  &
                     ComputeBackgroundMode1, & 
                     ComputeBackgroundMode2, &
                     ComputeBackgroundMode3, &
                     ComputeBackgroundMode4
end interface

real(double), private :: poissonProbability, backgroundMean

contains
!*****************************************
! Function BackgroundMode
!*****************************************

function BackgroundMode1(nData, data, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
 

    real(single), pointer :: list(:) => null()
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k, allocErr
     
    real(single)   :: mean, median
    real(double)   :: variance, testSigma
    real(single)   :: highValue, lowValue
    integer(int32) :: m, ndataMin
    
    real(single) :: nsigma1   


    nList = nData
    nDataMin = nData / 2
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data so the median can be
    ! quickly computed

    call HeapSort(list, nList)

    nsigma1 = nsigma
  
     ! reject data at more than nSigma from data
   
    testSigma = 0.0

    sigmaClip: do i = 1, 20   !  was 50



      
        mean = sum( list(1 : nList)) / dble(nList)    
        median = SortedMedian(list, nlist)

        if (median .le. 2.0 .and. (mean - median) .le. 1.0) then
            nsigma1 = 3.0
            median = mean
        end if

                
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
        variance = sum(dble(list(1 : nlist) - median) ** 2)

        variance = variance / dble(nlist - 1)
        sigma = sqrt(variance) 

         
        if (sigma .le. sqrt(median)) nsigma1 = 3.0
        sigma = max(1.0, sigma)
     
        lowValue = max(0.0, median - 5.0 * sigma)
        highValue = median + nSigma1 * sigma
    
        m = ishft(nlist, -1) 
        do k = m + 1, nlist
            if(list(k) .ge. lowValue .and. list(k) .le. highValue) then

                m = m + 1            
            else
                exit   
            end if
        end do
       

        if(m .lt. ndataMin .or. sigma .lt. 0.5) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip




    mode = sum( list(1 : nList)) / dble(nList)    


    variance = sum(dble(list(1 : nlist) - mode) ** 2)
    variance = variance / dble(nlist - 1)
    sigma = sqrt(variance)



    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")




end function BackgroundMode1




!*****************************************
! Function BackgroundMode
!*****************************************
function BackgroundMode2(nData, data, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

    real(single), intent(out) :: sigma
    

    real(single), pointer :: list(:)
   
   
    real(single)   :: median, mode, testMedian
    real(single) :: adev, mean, sdev, skew, var
  
    integer(int32) :: n1, nMin
!median = 20.0
!return

    nMin = ishft(nData, -1)

    n1 = nData
    list => AllocateSingle(1, nData, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data
    call HeapSort(list, nData)
    testMedian = 0.0
    do while (n1 > nMin)
        
        
        median = SortedMedian(list, n1)
        if (median .lt. 0.1) exit
        if (abs(median - testMedian) / median .lt. 0.0001) exit
!        f = max(0.1, sqrt(median))
        call SMoments(nData, data, mean, adev, sdev, skew, var)


!        if ( abs(adev) < f) exit
         
        n1 = n1 - 1
        testMedian = median       
    end do
    

    deallocate(list)
    
    mode = 2.5 * median - 1.5 * mean
    if(mode .lt. 0.0) mode = mean

!write(*, *) "Background mode=", mode, "+/-", sigma, " n=", nlist, " no of iterations=", i  
    sigma = sdev        
end function BackgroundMode2





!*****************************************
! Function BackgroundMode3
!*****************************************
function BackgroundMode3(nData, data, &
                        tolerance, sigma) result(mode)

    integer(int32), intent(in) :: nData
    real(single) :: data(:)

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(out) :: sigma
 

    real(single), pointer :: list(:)
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k, allocErr
     
    real(single)   :: testMedian, mean, median
    real(double)   :: sqrtBackgroundMean, fmin
 
    integer(int32) :: m, ndataMin

    real(double) :: lowValue, highValue   

    integer, save :: count = 0
    nList = nData
    nDataMin = nData / 2
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data so the median can be
    ! quickly computed
!    call HeapSort(list, nList)
             
  
    ! reject data at more than nSigma from data
   
    testMedian = 0.0
     sigmaClip: do i = 1, 20  
      
             
        median = SortedMedian(list, nlist)

                
        !*******************************
        ! Now do sigma clipping
        !*******************************
        backgroundMean = dble(median)
        sqrtBackgroundMean = 6.0_double * sqrt(backgroundMean)

        poissonProbability = 0.00135_double
  poissonProbability = 0.01_double


        fmin = brent( max(0.0_double, backgroundMean - sqrtBackgroundMean), backgroundMean, &
                   1.0d-6, lowValue, backgroundFunction)

        poissonProbability = 1.0_double - poissonProbability

        fmin = brent(backgroundMean, backgroundMean + max(5.0_double, backgroundMean + sqrtbackgroundMean),  &
                 1.0d-6, highValue, backgroundFunction)



if (count .lt. 100) then
count = count + 1

write(*, *) "Iteration ", i, " median=", median, " old median=", testMedian
write(*, *) "low value=", lowValue, " high value=", highvalue
end if

        m = ishft(nlist, -1) 

        do k = m + 1, nlist

!            if(list(k) .le. highValue .and. list(k) .ge. lowValue) then
            if(list(k) .le. highValue) then

 
                m = m + 1
            
            else
                exit   
            end if
        end do
       

        if(m .lt. nDataMin) exit sigmaClip
     
        if(abs((median - testMedian) / median) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testMedian = median
        
    end do sigmaClip


    mean = sum( list(1 : nList)) / dble(nList)    

    mode = 2.5 * median - 1.5 * mean


if (count .lt. 100) then
write(*, *) "mean=", mean, " mode=", mode, " median=", median
end if
    if(mode .lt. 0.0) mode = mean

mode = mean
    !********************************
    sigma = sum(dble(list(1 : nlist) - mode) ** 2)

    sigma = sqrt(sigma / dble(nlist - 1))
 
    

    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")



end function BackgroundMode3
!***********************************************************
!
!***********************************************************
real(double) function backgroundFunction(x)
    real(double), intent(in) :: x

    backgroundFunction = abs(poissonProbability - gammap(backgroundMean, x))




end function backgroundFunction

!**********************************************************************
!
!********************************************************************** 
subroutine SMoments(n, data, mean, adev, sdev, skew, var)
    integer(int32), intent(in) :: n
    real(single), intent(in) :: data(1 : n)
    real(single), intent(out) :: mean, adev, sdev, skew, var   
    real(double) :: mean1, adev1, sdev1, skew1, var1
    real(double) :: ep, p, s
    integer :: i

    mean1 = sum(dble(data)) / dble(n)
    adev1 = 0.0_double
    sdev1 = 0.0_double
    skew1 = 0.0_double
    var1 = 0.0_double
    ep = 0.0_double

    do i = 1, n
        s = dble(data(i)) - mean
        ep = ep + s
        adev1 = adev1 + s
        p = s * s
        var1 = var1 + p
        skew1 = skew1 * p
    end do

    adev = real(adev1 / dble(n))
    var1 = (var1 - ep * ep / dble(n)) / dble(n - 1)
    sdev1 = sqrt(var1)
    if (var1 == 0.0_double) then
      
        skew = 0.0
    else
        skew = real(skew1 / (dble(n) * var1 * sdev1))
    end if
    sdev = real(sdev1)
    var = real(var1)
    mean = real(mean1)
end subroutine SMoments
!*****************************************
! Function BackgroundMode
!*****************************************
function BackgroundModeOrig(nData, data, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)
    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
    real(single), pointer :: list(:)
    integer(int32) :: nlist
   
    real(double)   :: mean
    real(single)    :: mode
    integer(int32) :: i, k

     
    real(single)   :: median
    real(double)   :: variance, originalSigma, testSigma
    real(single)   :: lowValue, highValue
    integer(int32) :: allocErr, m

   

    nList = nData
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    !**************************************
    ! Sort the data so the median can be
    ! quickly computed
    !**************************************
    call HeapSort(list, nList)

             
    mode = SortedMedian(list, nList)
      
    !**************************************
    ! Get variance and std deviation
    !**************************************
    variance = 0.0_double
    do k = 1, nList 
        variance = variance + dble((list(k) - mode) ** 2)
    end do 
    variance = variance / real(nList - 1)
    
    
    originalSigma = sqrt(variance)
    lowValue = mode - nSigma * originalSigma

    highValue = mode + nSigma * originalSigma


    !**********************************************
    ! reject data at more than nSigma from data
    !**********************************************
    testSigma = originalSigma

    sigmaClip: do i = 1, 20   !  was 50

       
        m = 0
        mean = 0.0_double
        
        do k = 1, nlist
            if(list(k) .ge. lowValue .and. list(k) .le. highValue) then

  
                m = m + 1
                list(m) = list(k)
                mean = mean + dble(list(m))
            end if
        end do
        nlist = m
        if(nlist .lt. 20) exit sigmaClip
       
        mean = mean / dble(nList)     
        median = SortedMedian(list, nlist)
        mode = 2.5 * median - 1.5 * mean
        if (mode .le. 0.0 .or. mode .gt. 2.0 * median) mode = median
        
 
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
        variance = 0.0_double
        do k = 1, nlist 
            variance = variance + dble((list(k) - mode) ** 2)
        end do 
        variance = variance / (nlist - 1)
 
        sigma = sqrt(variance)

     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     
        lowValue = mode - nSigma * sigma
        
        highValue = mode + nSigma * sigma
     
   
        testSigma = sigma
        
    end do sigmaClip


    if(abs((sigma - originalSigma) / sigma) .gt. tolerance) then
        median = SortedMedian(list, nlist)
        mode = 2.5 * median - 1.5 * mean
    else
        mode = mean
    end if
    if (mode .le. 0.0 .or. mode .gt. 2.0 * median) mode = mean

    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
!write(*, *) "Background mode=", mode, "+/-", sigma, " n=", nlist, " no of iterations=", i  
    
end function BackgroundModeOrig

!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode(image, levelImage, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)

    integer(int16), pointer   :: levelImage(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j

    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer :: nx1
   
    nx1 = nx - 1
    list => AllocateSingle(1, int(nx * ny, int32), "list (Computemode)")
    
 
    nlist = 0
 
    do j = 0, ny - 1
        do i = 0, nx1

            if (levelImage(i, j) .eq. 0 .and. image(i, j) .ge. 0.0) then
                nList = nList + 1
                list(nlist) = image(i, j) 
               
            end if
        end do
    end do 


   

    mode = BackgroundMode(nList, list, 0.2, nSigma, sigma)  ! 

!write(*, *) "omskybackground_mod.f90 line 372 mode=", mode, " nList=", nList
 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode

!**************************************************
! ComputeBackgroundMode1
! Computes the background mode counts
! using the image (background + source) 
!***************************************************
     
 
function ComputeBackgroundMode1(size, background, &
             nSigma, standardDeviation) result(mode)

    integer(int32), intent(in)  :: size
    real(single), pointer :: background(:)

    real(single), intent(in)    :: nSigma
    real(single), intent(out)   :: standardDeviation

    real(single)                :: mode
                   
    mode = BackgroundMode(size, background, 0.2, nSigma, standardDeviation) ! ws 0.1
    
end function ComputeBackgroundMode1
!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode2(image, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j

    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer :: nx1
   
    nx1 = nx - 1
    list => AllocateSingle(1, int(nx * ny, int32), "list (Computemode)")
    
 
    nlist = 0
 
    do j = 0, ny - 1
        do i = 0, nx1

            if (image(i, j) .ge. 0.0) then
                nList = nList + 1
                list(nlist) = image(i, j) 
            end if  
        end do
    end do 
   
    mode = BackgroundMode(nList, list, 0.2, nSigma, sigma)  ! 

 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode2

!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode3(image, sigma) result(mode)

     real(single) :: image(:, :)
 
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j

    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer(int32) :: nx, ny
     
    nx = size(image, 1)
    ny = size(image, 2)


    list => AllocateSingle(1, nx * ny, "list (Computemode)")
  
 
    nlist = 0


!write(*, *) "Computebackgroundmode3 line 666"
!write(*, *) "lbound image, 1=", lbound(image, 1), " ubound=", ubound(image, 1)
!write(*, *) "lbound image, 2=", lbound(image, 2), " ubound=", ubound(image, 2)

 
    do j = lbound(image, 2), ubound(image, 2)
        do i = lbound(image, 1), ubound(image, 1)

            if (image(i, j) .ge. 0.0) then           
                nList = nList + 1
                list(nlist) = image(i, j) 
            end if  
        end do
    end do 
!   write(*, *) "Line 680 nlist=", nlist, "nx=", nx, " ny=", ny
!write(*, *) "minval(list)=", minval(list), " max=", maxval(list)
 

    mode = BackgroundMode(nList, list, 0.2, 2.0, sigma)  ! 


 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode3


!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode4(image, levelImage, qualityImage, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)
    
    integer(int16), pointer :: levelImage(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j, nx1

    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
   
   
    list => AllocateSingle(1, int(nx * ny, int32), "list (Computemode)")
    
 
    nlist = 0
    nx1 = nx - 1
    do j = lbound(image, 2), ubound(image, 2)
        do i = lbound(image, 1), ubound(image, 1)

            if (levelImage(i, j) .eq. 0 .and. image(i, j) &
              .ge. 0.0 .and. .not. qualityImage(i, j)) then
                nList = nList + 1
                list(nlist) = image(i, j) 
               
            end if
        end do
    end do 


   
    mode = BackgroundMode(nList, list, 0.2, nSigma, sigma)  ! 
!write(*, *) "omskybackground_mod.f90 line 545 mode=", mode, " nList=", nList
 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode4


!********************************************************************
! returns an image. All pixels that belong to the same source have
! the same contents (source id)
! Modifies the level image
! If number > 0 pixels set to number
!********************************************************************
function CreateLevelImageMap(nx, ny, sources) result(levelImage)
    integer(int16), intent(in) :: nx, ny
    type(SourceParameters), pointer :: sources(:)
    integer(int16),          pointer    :: levelImage(:, :)   
    real(single)   :: sourceRadius, radius, x0, y0, xPixel, yPixel
    integer(int32)        :: r, i, j, k, i0, j0, i1, i2, j1, j2
    
    
    sourceRadius = real(10 / binX)


    r = ceiling(sourceRadius)    
    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "levelImage")
    if (.not. associated(sources)) return

    do k = 1, size(sources)
        if (sources(k)%extensionFlag .eq. 1) cycle
       

  
        x0 = sources(k)%xPos
        y0 = sources(k)%yPos
        i0 = int(x0)
        j0 = int(y0)


        ! Determine upper and lower array indices for summing counts
        i1 = i0 - r - 1
        i1 = max(0, i1)

        i2 = i0 + r + 1
        i2 = min(nx - 1, i2)

        j1 = j0 - r - 1
        j1 = max(0, j1)

        j2 = j0 + r + 1
        j2 = min(ny - 1, j2)

        do i = i1, i2
            xPixel = int(i)
            do j = j1, j2               
                yPixel = int(j)
                radius = sqrt((xPixel - x0) ** 2 + (yPixel - y0) ** 2)

                if (radius .le. sourceRadius) then
                    levelImage(i, j) = k
                end if

            end do
        end do

    end do
   
end function CreateLevelImageMap





subroutine RefineMode(ndata, data, mean, sigma)
    integer, intent(in) :: ndata
    real(single) :: data(:)
    real(single), intent(in) :: mean, sigma
    integer(int32) :: frequency(1 : 3)
    real(single) ::xmin, xmax, binWidth, refinedMode    
    integer(int32) :: i, k, noBins
    real(single) :: bin(1 : 3)

  


    noBins = 3
    binWidth = sigma * 2.6 * (0.48 + 6.2 / real(ndata)) ** 2
    
    xmin = mean - binwidth
    xmax = mean + binwidth
    frequency = 0
    do i = 1, ndata
        if (data(i) .lt. xmin .or. data(i) .gt. xmax) cycle
        
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1
        k= max(1, k)
        k = min(nobins, k)
        frequency(k) = frequency(k) + 1
 
    end do

    do i = 1, noBins
        bin(i) = xmin + real(i - 1) + binwidth / 2.0
    end do

    refinedMode = mean + (binwidth / 2.0) * real(frequency(1) - frequency(3)) / &
           real(frequency(1) - 2 * frequency(2) + frequency(3))

   


end subroutine RefineMode



end module Omskybackground_mod









