program testascregion

use dal
use ascregion_mod

implicit none
real :: x, y, xsize, ysize
real :: r1, r2,r3,r4,a1,a2
type(DataSetT) :: regDataSet
type(DataSetT) :: evDataSet
type(TableT) :: tableSet
type(ColumnT) :: col
real(kind=single), pointer :: vals(:)
integer :: i
integer, parameter :: maxrows = 500000
integer, parameter :: nPoly = 20
real :: xpoly(nPoly),ypoly(nPoly)
real :: xcen, ycen
call random_seed

regDataSet = createRegion("testregion.fts")

call getRandom(xcen, ycen, xsize, ysize, r1, r2, r3, r4, a1, a2)
do i = 1 , nPoly
   a1 = (real(i-1)/real(nPoly-1))*3.141592654*2.
   call random_number(r1)
   xpoly(i) = xcen + r1 * xsize*cos(a1)
   ypoly(i) = ycen + r1 * ysize*sin(a1)
enddo

call addPolygon(regDataSet, xpoly, ypoly, npoly, 16, .true.)


call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addPoint(regDataSet, x, y, 1, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addCircle(regDataSet, x, y, r1, 2, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addEllipse(regDataSet, x, y, r1, r2, a1, 3, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addAnnulus(regDataSet, x, y, r1, r2, 4, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addElliptAnnulus(regDataSet, x, y, r1, r1, r3, r4, a1, a2, 5, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addBox(regDataSet, x, y, xsize, ysize, 6, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRotbox(regDataSet, x, y, xsize, ysize, a1, 7, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRectangle(regDataSet, x, xsize, y, ysize, 8, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRotRectangle(regDataSet, x, xsize, y, ysize, a1, 9, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addPie(regDataSet, x, y, a1, a2, 10, .true.)
call addCircle(regDataSet, x, y, r1, 10, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addSector(regDataSet, x, y, a1, a2, 11, .true.)
call addCircle(regDataSet, x, y, r1, 11, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addDiamond(regDataSet, x, y, xsize, ysize, 12, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRhombus(regDataSet, x, y, xsize, ysize, 13, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRotDiamond(regDataSet, x, y, xsize, ysize, a1, 14, .true.)

call getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
call addRotRhombus(regDataSet, x, y, xsize, ysize, a1, 15, .true.)

call closeRegion(regDataSet)

evDataSet = dataSet("testeventlist.fts", create)
tableSet = addTable(evDataSet,"EVENT LIST", maxrows)
col = addColumn(tableSet, "X", real32)
vals => real32data(col)
call random_number(vals)
vals = vals * 1000.
call release(col)
col = addColumn(tableSet, "Y", real32)
vals => real32data(col)
call random_number(vals)
vals = vals * 1000.
call release(col)
call release(evDataSet)

end program testascregion

subroutine getRandom(x, y, xsize, ysize, r1, r2, r3, r4, a1, a2)
implicit none

real, parameter :: xMax =  1000., yMax = 1000.
real, parameter :: xMaxSize = 100., yMaxSize = 100., rMax = 10.
real :: x, y
real :: xsize, ysize
real :: r1, r2, r3, r4
real :: a1, a2

call random_number(x)
x = x * xMax

call random_number(y)
y = y * yMax

call random_number(xsize)
xsize = xsize * xMaxSize
xsize = max(xsize, 20.)

call random_number(ysize)
ysize = ysize * yMaxSize
ysize = max(ysize, 20.)

call random_number(r1)
r1 = r1 * rMax/10.
r1 = max(r1,20.)

call random_number(r2)
r2 = r1 * (1. + r2)

call random_number(r3)
r3 = r2 * (1. + r3)

call random_number(r4)
r4 = r3 * (1. + r4)

call random_number(a1)
a1 = a1 * 360.

call random_number(a2)
a2 = a2 * 360.
end subroutine getRandom


